"""
CopyRight (c) 2024 - Chris Laffra - All Rights Reserved.

Tests the `intercept_last_expression` function from the `api` module, which is responsible
for intercepting and modifying the last expression in a given script.
"""

import sys
import unittest

sys.path.append("..")

from tests import mocks # pylint: disable=wrong-import-position,unused-import
from static import api # pylint: disable=wrong-import-position


class TestLastExpression(unittest.TestCase):
    """
    Tests the `intercept_last_expression` function from the `api` module, which is responsible
    for intercepting and modifying the last expression in a given script.
    """

    def test_intercept_empty(self):
        """
        Tests that `intercept_last_expression` correctly handles an empty script.
        """
        script = ''
        actual = api.intercept_last_expression(script)
        expected = ''
        self.assertEqual(actual, expected)

    def test_intercept_multiline_dict(self):
        """
        Tests that `intercept_last_expression` correctly handles a multiline dictionary expression.
        """
        script = "x = {\n    'y':\n     4,\n }"
        actual = api.intercept_last_expression(script)
        expected = "_ = " + script
        self.assertEqual(actual, expected)

    def test_intercept_multiline_call(self):
        """
        Tests that `intercept_last_expression` correctly handles a multiline function call expression.
        """
        script = "print(\n1,\n2,\n3\n)\nprint(\n4,\n3,\n4\n)"
        actual = api.intercept_last_expression(script)
        expected = "print(\n1,\n2,\n3\n)\n_ = print(\n4,\n3,\n4\n)"
        self.assertEqual(actual, expected)

    def test_intercept_multiline(self):
        """
        Tests that `intercept_last_expression` correctly handles a script with multiple lines of assignments.
        """
        script = "x = 1\ny = 2\nz = 3"
        actual = api.intercept_last_expression(script)
        expected = "x = 1\ny = 2\n_ = z = 3"
        self.assertEqual(actual, expected)



"""
Copyright (c) 2024 laffra - All Rights Reserved. 

Manages the preview views for a sheet in the application.
"""

import constants
import history
import ltk
import models
import selection

previews = {}


class PreviewView(ltk.Div):
    """
    Represents a preview view widget for a sheet in the application.
    """
    def __init__(self, sheet: models.Sheet, preview: models.Preview):
        super().__init__()
        self.sheet = sheet
        self.model = preview
        previews[self.model.key] = self
        try:
            offset = ltk.find(f"#{self.model.key}").offset()
            left, top = offset.left, offset.top
        except AttributeError:
            left, top = 100, 100
        dx = 80 if len(previews) % 2 == 0 else 40
        ltk.find(f".preview-{self.model.key}").remove()
        ltk.find(".sheet-grid").append(self
            .addClass("preview")
            .addClass(f"preview-{self.model.key}")
            .attr("id", f"preview-{self.model.key}")
            .css("position", "absolute")
            .css("left", self.model.left or left + self.width() + dx)
            .css("top", self.model.top or top)
            .css("width", self.model.width or "fit-content")
            .css("height", self.model.height or "fit-content")
            .on("click", ltk.proxy(lambda event: self.click()))
            .on("mousemove", ltk.proxy(lambda event: self.move()))
            .on("mouseleave", ltk.proxy(lambda event: selection.remove_arrows(0)))
            .on("resizestop", ltk.proxy(lambda event, ui: self.resize(event)))
            .on("dragstop", ltk.proxy(lambda event, ui: self.dragstop(event)))
            .draggable(ltk.to_js({ "containment": ".sheet", "handle": ".preview-header" }))
        )
        self.model.listen(self.model_changed)
        self.set_html(self.model.html)

    def model_changed(self, preview, info): # pylint: disable=unused-argument
        """
        Updates the CSS properties of the PreviewView widget based on changes to the corresponding 
        properties in the Preview model.
        
        The `model_changed` method is called whenever the `Preview` model associated with this `PreviewView` 
        instance is updated. It takes the name of the updated property (`name`) and the new value 
        (`getattr(self.model, name)`), and applies the new value to the corresponding CSS property of 
        the `PreviewView` widget.
        """
        name = info["name"] # values are left, top, width, and height
        self.css(name, getattr(self.model, name))

    def dragstop(self, *args): # pylint: disable=unused-argument
        """
        Event handler for when the user stops dragging it.
        """
        history.add(
            models.PreviewPositionChanged(
                self.model.key, self.model.left, self.model.top, self.css("left"), self.css("top")
            ).apply(self.sheet.model)
        )

    def resize(self, event, *args): # pylint: disable=unused-argument
        """
        Event handler for when the user resizes the preview.
        """
        ltk.find(event.target) \
            .find("img, iframe") \
            .css("width", "100%") \
            .css("height", f"calc(100% - {constants.PREVIEW_HEADER_HEIGHT})")
        history.add(models.PreviewDimensionChanged(
            self.model.key, self.css("width"), self.css("height")).apply(self.sheet.model)
        )

    def move(self):
        """
        Event handler for when the user moves the preview by dragging the header.
        """
        self.draw_arrows()

    def click(self):
        """
        Event handler for when the user click inside the preview header.
        """
        self.appendTo(self.parent())  # raise

    def toggle_size(self, event):
        """
        Event handler for when the user toggles the size of the preview widget between a minimized and expanded state.
        """
        minimize = self.height() > constants.PREVIEW_HEADER_HEIGHT
        height = constants.PREVIEW_HEADER_HEIGHT if minimize else "fit-content"
        width = constants.PREVIEW_HEADER_WIDTH if minimize else "fit-content"
        self \
            .attr("minimized", minimize) \
            .height(height) \
            .width(width) \
            .find(".ui-resizable-handle") \
                .css("display", "none" if minimize else "block")
        ltk.find(event.target).text("+" if minimize else "-")

    def draw_arrows(self):
        """
        Draws an arrow pointing to the preview key.
        """
        selection.remove_arrows(0)
        self.sheet.get_cell(self.model.key).draw_arrows([])
        ltk.window.addArrow(ltk.find(f"#{self.model.key}"), self.element.find(".preview-key"), self.model.key)

    def set_html(self, html):
        """
        Sets the HTML content of the preview widget and updates the UI accordingly.
        """
        self.model.html = self.fix_html(html)
        toggle_size_label = "-" if self.height() > constants.PREVIEW_HEADER_HEIGHT or self.height() == 0 else "+"
        self.empty().append(
            ltk.HBox(
                ltk.Text(self.model.key).addClass("preview-key"),
                ltk.Button(toggle_size_label, ltk.proxy(lambda event: self.toggle_size(event))).addClass("toggle") # pylint: disable=unnecessary-lambda
            ).addClass("preview-header"),
            ltk.Div(
                ltk.create(self.model.html)
            ).addClass("preview-content")
        )
        self.make_resizable()
        self.draw_arrows()
        self.fix_images()

    def fix_html(self, html):
        """
        Fixes the HTML content of the preview widget to ensure it is properly formatted.
        
        Args:
            html (str): The HTML content to be fixed.
        
        Returns:
            str: The fixed HTML content.
        """
        html = html.replace("script src=", "script crossorigin='anonymous' src=")
        if not html.startswith("<"):
            html = f"<div>{html}</div>"
        return html

    def fix_images(self):
        """
        Fixes the `src` attribute of all `img` elements in the preview content
        to include the `crossorigin="anonymous"` attribute.
        """
        script = """
            $("img").each(function () {
                $(this).attr("crossorigin", "anonymous").attr("src", $(this).attr("src"))
            });
        """
        self.find("iframe").contents().find("body").append(
            ltk.create("<script>").html(script)
        )

    def make_resizable(self):
        """
        Makes the preview widget resizable using jQuery UI `resizable`.
        """
        preview = ltk.find(f"#preview-{self.model.key}")
        preview.find("img, iframe") \
            .css("width", "100%") \
            .css("height", f"calc(100% - {constants.PREVIEW_HEADER_HEIGHT})")
        preview.resizable(ltk.to_js({"handles": "se"}))
        display = "block" if preview.height() > constants.PREVIEW_HEADER_HEIGHT else "none"
        preview.find(".ui-resizable-handle") \
            .css("display", display)


def load(sheet):
    """
    Loads the preview views for each key in the sheet's model previews dictionary.
    
    Args:
        sheet (Sheet): The sheet object containing the previews.
    """
    for preview in sheet.model.previews.values():
        PreviewView(sheet, preview)


def add(sheet, key, html):
    """
    Adds a new preview view to the previews dictionary, or updates an existing one.
    
    Args:
        sheet (Sheet): The sheet object containing the previews.
        key (str): The key associated with the preview.
        html (str): The HTML content to be displayed in the preview.
    """
    if not html or html == "None":
        return
    if not key in previews:
        PreviewView(sheet, sheet.model.get_preview(key, html=html))
        old_html = ""
    else:
        old_html = previews[key].model.html
    previews[key].set_html(html)
    if old_html != html:
        history.add(models.PreviewValueChanged(key, html))

def remove(key):
    """
    Removes a preview view from the previews dictionary.
    
    Args:
        key (str): The key associated with the preview to be removed.
    """
    if key in previews:
        del previews[key]



"""
CopyRight (c) 2024 - Chris Laffra - All Rights Reserved.

This module provides a worker for a PyScript-based application to run Python
code, find dependencies, and perform code completion.
"""

import json
import sys
import time
import re
import traceback

import requests

import api
import lsp
import constants
import ltk

import js # type: ignore    pylint: disable=import-error
import polyscript # type: ignore    pylint: disable=import-error
import pyodide # type: ignore    pylint: disable=import-error
import pyscript # type: ignore    pylint: disable=import-error

window = pyscript.window
get = window.get
cache = { }
results = { }
OriginalSession = requests.Session
inputs_cache = {}


class PyScriptResponse():
    """
    Represents a response from a PyScript-based HTTP request.
    
    This class encapsulates the response data from a PyScript-based HTTP request,
    including the URL, status code, and response content.
    
    Attributes:
        url (str): The URL of the request.
        status (int): The HTTP status code of the response.
        content (str): The content of the response.
    
    Methods:
        json(): Parses the response content as JSON and returns the resulting object.
        text(): Returns the response content as a string.
    """
    def __init__(self, url, status, content):
        self.url = url
        self.status = status
        self.content = content

    def json(self):
        """
        Parses the response content as JSON and returns the resulting object.
        """
        return json.loads(self.content)

    def text(self):
        """
        Returns the response content as a string.
        """
        return self.content

    def __repr__(self):
        return f"Response[{self.url}, {self.status}, {self.content[32]}...]"


class PyScriptSession(OriginalSession):
    """
    Represents a custom session class that overrides the default `requests.Session`
    class to handle HTTP requests using PyScript's `window.XMLHttpRequest`.
    
    This class is used to make HTTP requests from within a PyScript-based worker environment,
    where the standard `requests` library may not be available or may not work as expected.
    
    The `request()` method of this class sends an HTTP request using the `window.XMLHttpRequest`
    object, with the URL modified to include the `constants.URL` parameter. The response
    is then wrapped in a `PyScriptResponse` object and returned.
    """

    def request( self, method, url, data=None, headers=None, **vargs):  #pylint: disable=arguments-differ,unused-argument
        xhr = window.XMLHttpRequest.new()
        xhr.open(method, f"load?{constants.URL}={url}", False)
        xhr.setRequestHeader("Authorization", (headers or self.headers).get("Authorization"))
        xhr.send(data)
        return PyScriptResponse(url, xhr.status, xhr.responseText)


requests.Session = PyScriptSession
requests.session = PyScriptSession



def setup():
    """
    Sets up the worker.
    """
    import builtins # pylint: disable=import-outside-toplevel
    if js.document is pyscript.document:
        return
    js.document = pyscript.document  # patch for matplotlib inside workers

    def worker_print(*args, file=None, end=""): # pylint: disable=unused-argument
        polyscript.xworker.sync.publish(
            "Worker",
            "Application",
            constants.TOPIC_WORKER_PRINT, f"[Worker] {' '.join(str(arg) for arg in args)}"
        )

    builtins.print = worker_print

completion_cache = {}

def get_image_data(figure):
    """
    Converts a Matplotlib figure to an HTML image representation.
    
    Args:
        figure (matplotlib.figure.Figure): The Matplotlib figure to convert.
    
    Returns:
        str: An HTML string representing the figure as an image.
    """
    from io import BytesIO # pylint: disable=import-outside-toplevel
    import base64 # pylint: disable=import-outside-toplevel

    #
    # Use the built-in agg backend.
    # This reduces rendering time for plots from ~6s to 70ms
    #
    matplotlib.use("agg")
    bytes_io = BytesIO()
    figure.set_edgecolor("#BBB")
    figure.savefig(bytes_io, bbox_inches="tight", format="png")
    bytes_io.seek(0)
    matplotlib.pyplot.close(figure)
    encoded = base64.b64encode(bytes_io.read())
    return f"""<img src="data:image/png;base64,{encoded.decode('utf-8')}">"""


def create_preview(result): # pylint: disable=too-many-return-statements
    """
    Creates a preview of the given result object.
    
    The preview can be in the form of an HTML representation, an image, or a
    string representation, depending on the type of the result object.
    
    Args:
        result: The result object to create a preview for.
    
    Returns:
        A string representation of the preview.
    """
    if str(result) == "DataFrame":
        return str(result)
    if "plotly" in str(type(result)):
        try:
            import plotly # pylint: disable=import-outside-toplevel
            html = plotly.io.to_html(result, default_width=500, default_height=500)
            return html
        except ImportError:
            pass
    try:
        return get_image_data(result)
    except Exception: # pylint: disable=broad-except
        pass  # print(traceback.format_exc())
    try:
        return get_image_data(result.get_figure())
    except Exception: # pylint: disable=broad-except
        pass  # print(traceback.format_exc())
    try:
        return result._repr_html_() # pylint: disable=protected-access
    except Exception: # pylint: disable=broad-except
        pass  # print(traceback.format_exc())
    try:
        return api.get_dict_table(result)
    except Exception: # pylint: disable=broad-except
        pass  # print(traceback.format_exc())
    return str(result)


def get_visualization_prompt(key, columns):
    """
    Generates a prompt for visualizing a dataframe as a matplotlib figure.
    
    Args:
        key (str): The key used to store the dataframe in the cache.
        columns (list[str]): The column names of the dataframe.
    
    Returns:
        str: A prompt that can be used to generate a visualization of the dataframe.
    """
    return f"""
Visualize a dataframe as a matplotlib figure in the code and call it "figure".
I already have it stored in a variable called "{key}".
Here are the column names for the dataframe:
{columns}
    """.strip()


def generate_completion(key, prompt):
    """
    Generates a completion for a given prompt using an external API.
    
    Args:
        key (str): A unique key to identify the completion request.
        prompt (str): The prompt to generate the completion for.
    
    Returns:
        None. The completion result is published to a message queue for further processing.
    """
    data = {constants.PROMPT: prompt}
    start = time.time()
    url = "complete"

    def success(response, status, xhr): # pylint: disable=unused-argument
        data = json.loads(window.JSON.stringify(response))
        if data.get(constants.STATUS, "ok") == "error":
            text = data[constants.ERROR]
        else:
            text = data["text"].replace("plt.show()", "figure")
            completion_cache[prompt] = data
        polyscript.xworker.sync.publish(
            "Worker",
            "Application",
            constants.TOPIC_WORKER_COMPLETION,
            json.dumps(
                {
                    "key": key,
                    "prompt": prompt,
                    "text": text,
                    "duration": time.time() - start,
                }
            ),
        )

    def error(data, status, xhr): # pylint: disable=unused-argument
        polyscript.xworker.sync.publish(
            "Worker",
            "Application",
            constants.TOPIC_WORKER_COMPLETION,
            json.dumps(
                {
                    "key": key,
                    "prompt": prompt,
                    "text": data,
                    "duration": time.time() - start,
                }
            ),
        )

    window.ltk_post(
        url,
        data,
        pyodide.ffi.create_proxy(success),
        None,
        pyodide.ffi.create_proxy(error),
    )


def run_in_worker(script):
    """
    Executes a Python script in the worker context, with access to the global
    cache and results dictionaries, as well as the pyodide, pyscript, and pysheets modules.
    
    The script is first intercepted to ensure the last expression is returned
    as the result. The script is then executed in the _globals environment,
    which includes the cache and results dictionaries, as well as the pyodide,
    pyscript, and pysheets modules. The result of the script execution is returned.
    """

    _globals = {}
    _globals.update(cache)
    _globals.update(results)
    _globals["pyodide"] = pyodide
    _globals["pyscript"] = pyscript
    _globals["pysheets"] = sys.modules["pysheets"] = api.PySheets(None, cache)
    _locals = _globals
    script = api.intercept_last_expression(script)
    # print("Executing script:", script)
    exec(script, _globals, _locals) # pylint: disable=exec-used
    return _locals["_"]


def run(data): # pylint: disable=too-many-locals
    """
    Executes a Python script in the worker context, with access to the global
    cache and results dictionaries, as well as the pyodide, pyscript, and pysheets modules.
    """
    start = time.time()
    key, script, inputs = data
    cache.update(inputs)
    try:
        result = run_in_worker(script)
    except Exception:  # pylint: disable=broad-except
        tb = traceback.format_exc()
        try:
            lines = tb.strip().split("\n")
            line = [line for line in lines if "<unknown>" in line][0]
            lineno = int(re.sub("[^0-9]", "", line))
            error = f"At line {lineno + 1}: {lines[-1]} - {tb}"
        except Exception as formatting_error: # pylint: disable=broad-except
            error = str(formatting_error)
            lineno = 1

        polyscript.xworker.sync.publish(
            "Worker",
            "Application",
            ltk.pubsub.TOPIC_WORKER_RESULT,
            json.dumps(
                {
                    "key": key,
                    "value": None,
                    "preview": "",
                    "duration": time.time() - start,
                    "lineno": lineno,
                    "error": error,
                    "traceback": tb,
                }
            ),
        )
        return

    try:
        kind = result.__class__.__name__
        cache[key] = results[key] = result
    except Exception:  # pylint: disable=broad-exception-caught
        polyscript.xworker.sync.publish(
            "Worker",
            "Application",
            ltk.pubsub.TOPIC_WORKER_RESULT,
            json.dumps(
                {
                    "key": key,
                    "script": script,
                    "value": None,
                    "preview": "",
                    "duration": time.time() - start,
                    "error": f"Worker result error: {type(error)}:{error}",
                    "traceback": traceback.format_exc(),
                }
            ),
        )
        return

    prompt = (
        get_visualization_prompt(key, results[key].columns.values)
        if kind == "DataFrame"
        else ""
    )

    try:
        preview = create_preview(result)
        base_kind = kind in ["int", "str", "float"]
        polyscript.xworker.sync.publish(
            "Worker",
            "Application",
            ltk.pubsub.TOPIC_WORKER_RESULT,
            json.dumps(
                {
                    "key": key,
                    "duration": time.time() - start,
                    "script": script,
                    "value": preview if base_kind else kind,
                    "preview": "" if base_kind else preview,
                    "prompt": prompt,
                    "error": (
                        preview
                        if kind == "str" and preview.startswith("ERROR:")
                        else None
                    ),
                }
            ),
        )
    except Exception as e: # pylint: disable=broad-exception-caught
        polyscript.xworker.sync.publish(
            "Worker",
            "Application",
            ltk.pubsub.TOPIC_WORKER_RESULT,
            json.dumps(
                {
                    "key": key,
                    "value": None,
                    "script": script,
                    "preview": "",
                    "duration": time.time() - start,
                    "error": f"Worker preview error: {type(e)}:{e}",
                    "traceback": traceback.format_exc(),
                }
            ),
        )


def handle_request(sender, topic, request): # pylint: disable=unused-argument
    """
    Handles various requests received by the worker process, including:
    - Generating code completions
    - Finding inputs for a given script
    - Running a script
    """

    try:
        data = json.loads(request)
        if topic == constants.TOPIC_WORKER_COMPLETE:
            generate_completion(data["key"], data["prompt"])
        elif topic == constants.TOPIC_WORKER_FIND_INPUTS:
            key, script = data["key"], data["script"]
            inputs = inputs_cache.get(script, None)
            try:
                if inputs is None:
                    inputs = api.find_inputs(script)
                inputs_cache[script] = inputs
                polyscript.xworker.sync.publish(
                    "Worker",
                    "Application",
                    constants.TOPIC_WORKER_INPUTS,
                    json.dumps(
                        {
                            "key": key,
                            "inputs": inputs,
                        }
                    ),
                )
            except Exception as e: # pylint: disable=broad-exception-caught
                polyscript.xworker.sync.publish(
                    "Worker",
                    "Application",
                    constants.TOPIC_WORKER_INPUTS,
                    json.dumps(
                        {
                            "key": key,
                            "error": str(e),
                            "inputs": [],
                        }
                    ),
                )
        elif topic == constants.TOPIC_WORKER_CODE_COMPLETE:
            text, line, ch = data
            completions = lsp.complete_python(text, line, ch, cache, results)
            polyscript.xworker.sync.publish(
                "Worker",
                "Application",
                constants.TOPIC_WORKER_CODE_COMPLETION,
                json.dumps(completions),
            )
        elif topic == ltk.pubsub.TOPIC_WORKER_RUN:
            run(data)
        else:
            print("Error: Unexpected topic request", topic)
    except Exception as e: # pylint: disable=broad-exception-caught
        print("Cannot handle request", request, e)


polyscript.xworker.sync.handler = handle_request

polyscript.xworker.sync.subscribe(
    "Worker", ltk.TOPIC_WORKER_RUN, "pyodide-worker"
)
polyscript.xworker.sync.subscribe(
    "Worker", constants.TOPIC_WORKER_COMPLETE, "pyodide-worker"
)
polyscript.xworker.sync.subscribe(
    "Worker", constants.TOPIC_WORKER_FIND_INPUTS, "pyodide-worker"
)
polyscript.xworker.sync.subscribe(
    "Worker", constants.TOPIC_WORKER_CODE_COMPLETE, "pyodide-worker"
)
polyscript.xworker.sync.publish(
    "Worker", "Sheet", ltk.pubsub.TOPIC_WORKER_READY, repr(sys.version)
)

try:
    import pandas               # pylint: disable=unused-import,wrong-import-position,import-error
    import matplotlib           # pylint: disable=wrong-import-position,import-error
    import matplotlib.pyplot    # pylint: disable=wrong-import-position,import-error
except (TypeError, ModuleNotFoundError, ImportError):
    pass # ignore load error on Github Actions build

setup()



"""
CopyRight (c) 2024 - Chris Laffra - All Rights Reserved.

This module contains unit tests for the `api.find_inputs()` function, which is
used to extract input variables from a given script.
"""

import json
import sys
import unittest
import unittest.mock

import ltk

sys.path.append("..")

from tests import mocks # pylint: disable=wrong-import-position,unused-import
from static import api # pylint: disable=wrong-import-position
from static import constants # pylint: disable=wrong-import-position
from static import worker # pylint: disable=wrong-import-position


class TestInputs(unittest.TestCase):
    """
    This class contains unit tests for the `api.find_inputs()` function, which is used
    to extract input variables from a given script. The tests cover various scenarios,
    including syntax errors, arithmetic operations, function calls, variable references,
    string handling, and multiple statements. The class also includes tests for the worker
    functionality, which mocks the polyscript subscribe and publish operations.
    """

    def check(self, script, expected):
        """
        Helper function for testing the `api.find_inputs()` function.
        """
        actual = sorted(api.find_inputs(script))
        self.assertEqual(set(actual), set(expected))

    def test_reference(self):
        """
        Simple expression containing a cell reference.
        """
        self.check("C2", ["C2"])

    def test_assignment(self):
        """
        Simple assignment with a cell reference as the right-hand side.
        """
        self.check("x = C2", ["C2"])

    def test_syntax_error(self):
        """
        Syntactically incorrect scripts have no inputs.
        """
        self.check("this causes a syntax error", [])

    def test_add(self):
        """
        Expression with an arithmetic operation containing two cell references.
        """
        self.check("C1+C2", ["C1", "C2"])

    def test_function_call(self):
        """
        Function call with a cell reference as an argument.
        """
        self.check("print(C1, 'Hello', C2, 'World')", ["C1", "C2"])

    def test_ignore_string(self):
        """
        Ignore cell references inside of strings.
        """
        self.check("x, y = 'C1', C2", ["C2"])

    def test_statements_1(self):
        """
        Multiple statements with a cell references in the first statement.
        """
        self.check("y = C2; x = 10", ["C2"])

    def test_statements_2(self):
        """
        Multiple statements with a cell references in the second statement.
        """
        self.check("x = 10\ny = C2", ["C2"])

    def test_assignment_to_cell_reference(self):
        """
        Ignore assignment to a cell reference.
        """
        self.check("C1 = 0\nx = C2\nC3 = 1", ["C2"])

    def test_function_range(self):
        """
        Detect cell reference ranges.
        """
        self.check("print('C1:C2')", ["C1", "C2"])

    def test_function_range_aa(self):
        """
        Detect cell reference ranges in the AA column.
        """
        self.check("print('AA1:AB4')", ["AB3", "AB1", "AA4", "AA3", "AA1", "AA2", "AB2", "AB4"])

    def mock_worker(self):
        """
        Mocks the worker functionality by subscribing to the 'Worker' topic for the
        'TOPIC_WORKER_FIND_INPUTS' event, and publishing to the 'Worker' topic when
        a request is received. This allows testing the worker functionality without
        actually running the worker as a web worker.
        """

        # mock polyscript subscribe
        def handle_worker_request(data):
            worker.handle_request("Worker", constants.TOPIC_WORKER_FIND_INPUTS, json.dumps(data))
        ltk.subscribe("Worker", constants.TOPIC_WORKER_FIND_INPUTS, handle_worker_request)

        # mock polyscript publish
        worker.publish = ltk.publish

    def check_worker(self, key, script, expected):
        """
        Checks the worker functionality by mocking the worker's behavior and verifying
        that the expected cell references are detected in the provided script.
        """
        self.mock_worker()
        handle_inputs = unittest.mock.Mock()

        # receive responses from the worker
        ltk.subscribe(
            "Worker",
            constants.TOPIC_WORKER_INPUTS,
            handle_inputs
        )

        # send a request to the worker
        ltk.publish(
            "Test",
            "Worker",
            constants.TOPIC_WORKER_FIND_INPUTS,
            {
                "key": key,
                "script": script,
            },
        )

        # verify the result
        handle_inputs.assert_called_once()
        data = json.loads(handle_inputs.call_args[0][0])
        self.assertEqual(data["key"], key)
        self.assertEqual(sorted(data["inputs"]), sorted(expected))

    @unittest.skip("Skipping worker tests as pubsub has changed")
    def test_worker_simple(self):
        """
        Checks the worker functionality by mocking the worker's behavior and verifying
        that the expected cell references are detected in the provided script.
        """
        self.check_worker("A1", "C1+C2", ["C1", "C2"])

    @unittest.skip("Skipping worker tests as pubsub has changed")
    def test_worker_range(self):
        """
        Checks the worker functionality by mocking the worker's behavior and verifying
        that the expected cell references are detected in the provided script.
        """
        self.check_worker("A1", "print('AA1:AB4')", ["AB3", "AB1", "AA4", "AA3", "AA1", "AA2", "AB2", "AB4"])



"""
CopyRight (c) 2024 - Chris Laffra - All Rights Reserved.

Profiles the execution of a Python program and generating a timeline visualization of the function calls.
"""

import sys
import time

import ltk
import state


COLORS = [
    "#FFE5E5", "#FFEFD5", "#FFFACD", "#E6E6FA", "#D8BFD8", "#DDA0DD", "#EEE9E9", "#F0FFF0",
    "#E0FFFF", "#FFDAB9", "#FFF0F5", "#FFE4E1", "#FFE4B5", "#D8BFD8", "#F5DEB3", "#F0E68C"
]
COLOR_MAP = {}


class Call():  # pylint: disable=too-few-public-methods
    """
    Represents a call in the execution timeline of a Python program.
    
    Attributes:
        when (float): The time in seconds when the call occurred.
        depth (int): The depth or nesting level of the call in the call stack.
        frame (frame): The Python frame object for the call.
        filename (str): The filename where the call occurred.
        lineno (int): The line number where the call occurred.
        duration (float): The duration of the call in seconds.
    """
    def __init__(self, when, depth, frame):
        self.when = when
        self.depth = depth
        self.frame = frame
        self.filename = frame.f_globals.get("__file__", "")
        self.lineno = frame.f_lineno
        self.duration = 0


class Span(ltk.Widget):
    """
    Represents a timespan for a call in the flame graph timeline visualization.
    """
    classes = [ "call" ]
    height = 20

    def __init__(self, when, call, pixels_per_second):
        ltk.Widget.__init__(self)
        module = call.frame.f_globals.get("__name__", "").replace("__main__", "pysheets")
        instance = call.frame.f_locals.get("self")
        clazz = ""
        if instance is not None:
            clazz = instance.__class__.__name__
            module = instance.__module__
        filename = call.filename.replace('/home/pyodide/', '')
        short_name = call.frame.f_code.co_name.replace("<", "&lt;")
        full_name = f"{module}.{clazz}.{short_name}"
        label = f"{short_name} | {call.duration:.2f}s at {when:.2f}s | {full_name}"
        self.html(f"<a href={self.get_vscode_link(filename, call.lineno)}>{label}</a>")
        self.attr("title", label)
        self.css("background-color", self.get_color(full_name))
        self.css("left", 8 + (call.when - when) * pixels_per_second)
        self.css("top", 40 + Span.height * call.depth)
        self.css("width", max(2, pixels_per_second * call.duration))

    def get_vscode_link(self, filename, lineno):
        """
        Returns a VSCode URL link for the specified filename and line number.
        
        Args:
            filename (str): The filename to include in the VSCode URL.
            lineno (int): The line number to include in the VSCode URL.
        
        Returns:
            str: A VSCode URL link for the specified filename and line number.
        """
        return f"vscode://file/{ltk.window.path}/{filename}:{lineno}"

    def get_color(self, name):
        """
        Returns a unique color for the given name.
        
        Args:
            name (str): The name to get a color for.
        
        Returns:
            str: The color for the given name, in CSS format (e.g. "#FFFFFF" for white).
        """
        if not name in COLOR_MAP:
            color = COLORS[hash(name) % len(COLORS)]
            COLOR_MAP[name] = color
        return COLOR_MAP[name]

    def __eq__(self, other):
        return isinstance(other, Span) and self.key == other.key


class FlameGraph(ltk.VBox):
    """
    A `FlameGraph` widget that displays a flame graph visualization of the profiled function calls.
    
    The `FlameGraph` class inherits from `ltk.VBox` and is responsible for rendering a flame graph
    visualization of the profiled function calls. It takes a list of `Call` objects as input and
    displays a flame graph with the function call durations and names.
    
    The flame graph can be zoomed in and out using the provided buttons, and the entire widget
    can be removed from the timeline container.
    """
    classes = [ "flame" ]

    def __init__(self, calls):
        ltk.VBox.__init__(self)
        self.calls = calls
        self.scale = 1
        call = self.calls[-1]
        name = call.frame.f_code.co_name.replace("<", "&lt;")

        def zoom(dy):
            self.scale = max(1, self.scale / 3) if dy < 0 else self.scale * 3
            self.render()

        def zoom_out(event): # pylint: disable=unused-argument
            self.scale = max(1, self.scale / 3)
            self.render()

        def zoom_in(event): # pylint: disable=unused-argument
            self.scale *= 3
            self.render()

        self.container = ltk.Div().on("wheel", lambda event: zoom(-event.originalEvent.deltaY))

        self.append(
            ltk.VBox(
                ltk.HBox(
                    ltk.Text(f"{name} took {call.duration}s").addClass("flame-title"),
                    ltk.Span("").css("flex", 1),
                    ltk.Button("-", lambda event: zoom_out(event)), # pylint: disable=unnecessary-lambda
                    ltk.Button("+", lambda event: zoom_in(event)), # pylint: disable=unnecessary-lambda
                    ltk.Button("x", lambda event: self.remove()).css("margin-right", 14), # pylint: disable=unnecessary-lambda
                ).addClass("flame-header"),
            ),
            self.container
        )
        self.render()

    def render(self):
        """
        Renders the flame graph visualization of the profiled function calls.
        """
        last = self.calls[-1]
        min_duration = 0 if self.scale > 1 else 0.01
        spans = [
            Span(last.when, call, self.scale * 300)
            for call in self.calls
            if call.duration > min_duration and not call.filename == "/home/pyodide/timeline.py"
        ]
        self.container.empty().append(*[span.element for span in spans])
        max_depth = max(call.depth for call in self.calls if call.duration > min_duration)
        self.css("width", max(span.width() + 20 for span in spans))
        self.css("height", 52 + (max_depth + 1) * Span.height)
        self.find(".flame-title").css("visibility", "visible" if self.scale > 1 else "hidden")


class Profiler():
    """
    The `Profiler` class is responsible for profiling the execution of a Python program and
    generating a flame graph visualization of the profiled function calls.
    """
    MIN_DURATION = 0.1

    calls = [] # when, duration, depth, frame
    stack = [] # when, frame

    def __init__(self):
        self.epoch = time.time()
        ltk.schedule(self.add_toggle, "add profiler toggle")
        self.enable_profile()

    def enabled(self):
        """
        Returns whether the timeline profiling is currently enabled.
        """
        return ltk.window.localStorage.getItem("timeline-enabled") == "true"

    def enable(self, enabled):
        """
        Enables or disables the timeline profiling functionality.
        
        Args:
            enabled (bool): True to enable the timeline profiling, False to disable it.
        """
        ltk.window.localStorage.setItem("timeline-enabled", enabled)

    def add_toggle(self):
        """
        Adds a toggle switch to the timeline container that enables or disables the timeline profiling functionality.
        """
        ltk.find(".timeline-container").append(
            ltk.Switch("enabled", self.enabled())
                .addClass("timeline-switch")
                .on("change", ltk.proxy(lambda event: self.toggle(ltk.find(event.target))))
        )

    def toggle(self, checkbox):
        """
        Toggles the profiling functionality of the Profiler class.
        
        Args:
            checkbox (ltk.Widget): The checkbox widget that represents the profiling toggle.
        """
        self.enable(checkbox.prop("checked"))
        self.enable_profile()

    def enable_profile(self):
        """
        Enables or disables the profiling functionality of the Profiler class.
        
        When profiling is enabled, the `sys.setprofile()` function is set to the `self.profile`
        method, which is responsible for recording function call information.
        When profiling is disabled, `sys.setprofile()` is set to `None` to stop the profiling.
        """
        sys.setprofile(self.profile if self.enabled() else None)

    def profile(self, frame, event, arg=None): # pylint: disable=unused-argument
        """
        Records function call information for the timeline profiling functionality.
        
        This method is set as the `sys.setprofile()` function when the timeline profiling is enabled.
        It is responsible for recording function call information, including the timestamp, call depth,
        and duration of each function call. The recorded information is stored in the `self.calls` list.
        
        If a function call takes longer than the `Profiler.MIN_DURATION` threshold, a console message is
        written with the function name and duration, and a FlameGraph is appended to the timeline container.
        
        Args:
            frame (frame): The current stack frame.
            event (str): The event type, either "call" or "return".
            arg (Any, optional): The argument passed to the function, if any.
        """
        now = round(time.time() - self.epoch, 3)
        if event == "call":
            Profiler.stack.append(Call(now, len(Profiler.stack), frame))

        if Profiler.stack and event == "return":
            call = Profiler.stack.pop()
            if call.depth < 13:
                call.duration = round(now - call.when, 3)
                self.calls.append(call)
                if call.depth == 0:
                    if call.filename != "/home/pyodide/timeline.py" and call.duration > Profiler.MIN_DURATION:
                        name = call.frame.f_code.co_name
                        state.console.write(
                            f"timeline-{name}", 
                            f"[Timeline] Call to '{name}' took {call.duration}s. See timeline. 🔔️"
                        )
                        ltk.find(".timeline-container").append(
                            FlameGraph(self.calls)
                        )
                    Profiler.calls = []

        return self.profile


def setup():
    """
    Enables the timeline profiling functionality by creating a new `Profiler` instance.
    
    This function is called to set up the timeline profiling feature. It checks if the
    `unittest` module is not loaded and if the `sys.setprofile()` function is available.
    If these conditions are met, it creates a new `Profiler` instance to start recording
    function call information for the timeline.
    """
    if 'unittest' not in sys.modules and hasattr(sys, "setprofile"):
        Profiler()
